-- If you want to use this mod for multiplayer, feel free to modify the source code so users won't be able to change variables.
-- To do this in BeamMP, you can the variables below local and then use functions & events to change them. For KissMP, you can do the same but events will be different.

local M = {}

local imgui = ui_imgui

M.enabled = false
M.decreaseToInitial = false
M.decrease = false
M.showUI = imgui.BoolPtr(false)
M.speed = 0.001
M.limitEnabled = false
M.limit = 0
M.ocean = nil
M.sky = nil
M.hideCoveredWater = true
M.floodWithRain = false
M.rainMultiplier = 0.01
M.rainVolume = 0
M.rainAmount = 0

local allWater = {}

local initialWaterPosition = nil
local oldMission = nil
local windowMinSize = 470

local ptrWaterLevel = imgui.FloatPtr(0)
local ptrHours = imgui.FloatPtr(0)
local ptrMinutes = imgui.FloatPtr(0)
local ptrSkyCol = ffi.new("float[4]", {0, 0, 0, 0})

-- functions
local function findObject(objectName, className)
    local obj = scenetree.findObject(objectName)
    if obj then return obj end
    if not className then return nil end

    local objects = scenetree.findClassObjects(className)
    for _, name in pairs(objects) do
        local object = scenetree.findObject(name)
        if string.find(name, objectName) then return object end
    end

    return
end

local function getRainAmount()
    local rain = findObject("rain_coverage","Precipitation")
    if rain then
        local drops = rain.numDrops
        return drops / 255 * M.rainMultiplier
    end

    return 0
end

local function tableToMatrix(tbl)
    local mat = MatrixF(true)
    mat:setColumn(0, tbl.c0)
    mat:setColumn(1, tbl.c1)
    mat:setColumn(2, tbl.c2)
    mat:setColumn(3, tbl.c3)
    return mat
end

local function setRainVolume(vol)
    local soundObj = findObject("rain_sound")
    if soundObj then
        soundObj:delete()
    end

    soundObj = createObject("SFXEmitter")
    soundObj.scale = Point3F(100, 100, 100)
    soundObj.fileName = String('/art/sound/environment/amb_rain_medium.ogg')
    soundObj.playOnAdd = true
    soundObj.isLooping = true
    soundObj.volume = vol
    soundObj.isStreaming = true
    soundObj.is3D = false
    soundObj:registerObject('rain_sound')
end

local function setup()
    M.enabled = false
    M.speed = 0.001
    M.limit = 0
    M.sky = nil
    M.limitEnabled = false
    M.decrease = false
    M.decreaseToInitial = false
    M.hideCoveredWater = true
    M.floodWithRain = false
    M.rainMultiplier = 0.01
    M.rainAmount = 0
    M.rainVolume = 0

    local gameTime = core_environment.getTimeOfDay().time
    local gameHours = math.floor(gameTime * 24)
    local gameMinutes = math.floor((gameTime * 24 - gameHours) * 60)
    ptrHours[0] = gameHours
    ptrMinutes[0] = gameMinutes

    setRainVolume(0)
    local rainObj = findObject("rain_coverage", "Precipitation")
    if rainObj then
        rainObj.numDrops = 0
    end

    M.ocean = nil
    allWater = {}
end

local function getOcean()
    local waterPlane = findObject("Ocean", "WaterPlane")
    if waterPlane then
        initialWaterPosition = waterPlane.position
    end

    return waterPlane
end

local function getSky()
    local scatterSky = findObject("sunsky", "ScatterSky")
    return scatterSky or nil
end

local function resetWaterLevel()
    if not M.ocean then return end
    M.ocean.position = initialWaterPosition
    ptrWaterLevel[0] = initialWaterPosition:getColumn(3).z
end

local hiddenWater = {}

local function resetAllWaterSources()
    for id, water in pairs(allWater) do
        water.isRenderEnabled = true
        hiddenWater[id] = false
    end
end

local function reset()
    resetWaterLevel()
    resetAllWaterSources()
end

local function getWaterLevel()
    if not M.ocean then return end
    return M.ocean.position:getColumn(3).z
end

local function setWaterLevel(level)
    if not M.ocean then log("W", "setWaterLevel", "M.ocean is nil") return end
    local c3 = M.ocean.position:getColumn(3)
    M.ocean.position = tableToMatrix({
        c0 = M.ocean.position:getColumn(0),
        c1 = M.ocean.position:getColumn(1),
        c2 = M.ocean.position:getColumn(2),
        c3 = vec3(c3.x, c3.y, level)
    })
end

local function getAllWater()
    local water = {}
    local toSearch = {
        "River",
        "WaterBlock"
    }

    for _, name in pairs(toSearch) do
        local objects = scenetree.findClassObjects(name)
        for _, id in pairs(objects) do
            if not tonumber(id) then
                local source = scenetree.findObject(id)
                if source then
                    table.insert(water, source)
                end
            else
                local source = scenetree.findObjectById(tonumber(id))
                if source then
                    table.insert(water, source)
                end
            end
        end
    end

    return water
end

local function hideCoveredWater()
    local height = ptrWaterLevel[0]

    for id, water in pairs(allWater) do
        local waterHeight = water.position:getColumn(3).z
        if M.hideCoveredWater and not hiddenWater[id] and waterHeight < height then
            water.isRenderEnabled = false
            hiddenWater[id] = true
        elseif waterHeight > height and hiddenWater[id] then
            water.isRenderEnabled = true
            hiddenWater[id] = false
        elseif not M.hideCoveredWater and hiddenWater[id] then
            water.isRenderEnabled = true
            hiddenWater[id] = false
        end
    end
end

local function tooltip(text)
    imgui.TextDisabled("(?)")
    if imgui.IsItemHovered() then
        imgui.BeginTooltip()
        imgui.PushTextWrapPos(imgui.GetFontSize() * 35.0);
        imgui.TextUnformatted(text)
        imgui.EndTooltip()
    end
end

local removeRainClicked = false
local windowWidth = 0

local function renderUI()
    floodui.pushStyle(imgui.StyleVar_WindowMinSize, imgui.ImVec2(330.0, windowMinSize))
    floodui.pushStyle(imgui.StyleVar_WindowTitleAlign, imgui.ImVec2(0.5, 0.5))
    floodui.pushStyle(imgui.StyleVar_WindowPadding, imgui.ImVec2(4.0, 4.0))
    floodui.pushStyle(imgui.StyleVar_ItemSpacing, imgui.ImVec2(6.0, 4.0))
    floodui.pushStyle(imgui.StyleVar_WindowBorderSize, 0.0)
    floodui.pushStyle(imgui.StyleVar_WindowRounding, 0.0)

    floodui.pushColor(imgui.Col_WindowBg, imgui.ImVec4(0.18, 0.2, 0.24, 1.0))

    floodui.pushColor(imgui.Col_TitleBg, imgui.ImVec4(0.18, 0.2, 0.24, 1.0))
    floodui.pushColor(imgui.Col_TitleBgActive, imgui.ImVec4(0.18, 0.2, 0.24, 1.0))
    floodui.pushColor(imgui.Col_TitleBgCollapsed, imgui.ImVec4(0.18, 0.2, 0.24, 1.0))

    floodui.pushColor(imgui.Col_Separator, imgui.ImVec4(0.85, 0.75, 0.55, 1.0))
    floodui.pushColor(imgui.Col_SeparatorHovered, imgui.ImVec4(0.85, 0.75, 0.55, 1.0))
    floodui.pushColor(imgui.Col_SeparatorActive, imgui.ImVec4(0.85, 0.75, 0.55, 1.0))

    floodui.pushColor(imgui.Col_CheckMark, imgui.ImVec4(0.85, 0.75, 0.55, 1.0))

    floodui.pushColor(imgui.Col_Button, imgui.ImVec4(0.37, 0.46, 0.56, 1.0))
    floodui.pushColor(imgui.Col_ButtonHovered, imgui.ImVec4(0.71, 0.56, 0.58, 1.0))
    floodui.pushColor(imgui.Col_ButtonActive, imgui.ImVec4(0.27, 0.36, 0.46, 1.0))

    floodui.pushColor(imgui.Col_FrameBg, imgui.ImVec4(0.37, 0.46, 0.56, 1.0))
    floodui.pushColor(imgui.Col_FrameBgHovered, imgui.ImVec4(0.71, 0.56, 0.58, 1.0))
    floodui.pushColor(imgui.Col_FrameBgActive, imgui.ImVec4(0.27, 0.36, 0.46, 1.0))

    floodui.pushColor(imgui.Col_SliderGrab, imgui.ImVec4(0.27, 0.36, 0.46, 1.0))
    floodui.pushColor(imgui.Col_SliderGrabActive, imgui.ImVec4(0.85, 0.75, 0.55, 1.0))

    if removeRainClicked then
        imgui.SetNextWindowSize(imgui.ImVec2(windowWidth, windowMinSize), imgui.Cond_Always)
        removeRainClicked = false
    end

    if imgui.Begin("Flood Controller", M.showUI, imgui.WindowFlags_NoDocking) then
        windowWidth = imgui.GetWindowWidth()

        local ptrEnabled = imgui.BoolPtr(M.enabled)
        local ptrDecreaseToInitial = imgui.BoolPtr(M.decreaseToInitial)
        local ptrDecrease = imgui.BoolPtr(M.decrease)
        local ptrLimitEnabled = imgui.BoolPtr(M.limitEnabled)
        local ptrFloodWithRain = imgui.BoolPtr(M.floodWithRain)
        local ptrHideCoveredWater = imgui.BoolPtr(M.hideCoveredWater)
        local ptrRainMultiplier = imgui.FloatPtr(M.rainMultiplier)
        local ptrSpeed = imgui.FloatPtr(M.speed)
        local ptrLimit = imgui.FloatPtr(M.limit)
        local ptrRainAmount = imgui.FloatPtr(0)
        local ptrRainVolume = imgui.FloatPtr(0)

        local rainObj = findObject("rain_coverage", "Precipitation")
        local soundObj = findObject("rain_sound")

        if rainObj then ptrRainAmount[0] = rainObj.numDrops end
        if soundObj then ptrRainVolume[0] = soundObj.volume end

        if imgui.Checkbox("Enabled", ptrEnabled) then
            M.enabled = ptrEnabled[0]
        end

        if imgui.Checkbox("Decrease to initial", ptrDecreaseToInitial) then
            M.decreaseToInitial = ptrDecreaseToInitial[0]
        end
        imgui.SameLine()
        tooltip("Avoid changing the speed, otherwise it won't perfectly match the initial level.\nIf you end up changing speed and see it's close to the initial position, lower the slider.")

        if imgui.Checkbox("Decrease", ptrDecrease) then
            M.decrease = ptrDecrease[0]
        end

        if imgui.Checkbox("Limit", ptrLimitEnabled) then
            M.limitEnabled = ptrLimitEnabled[0]
        end

        if imgui.Checkbox("Hide water sources under ocean", ptrHideCoveredWater) then
            M.hideCoveredWater = ptrHideCoveredWater[0]
            hideCoveredWater()
        end
        imgui.SameLine()
        tooltip("This will hide all water sources under the ocean level, if this is disabled you will see overlapped water sources.\nExample: the ocean is over a river so the river is now hidden so you can't see it.")

        if imgui.Checkbox("Flood with rain", ptrFloodWithRain) then
            M.floodWithRain = ptrFloodWithRain[0]
        end

        imgui.SameLine()
        tooltip("This will flood when there's rain.\nBy default, the multiplier is 0.0001, change this via flood.rainMultiplier.\nIt finds rain by checking for \"Precipitation\" object.\nIf you only want it to flood with rain, make sure speed is 0.")

        imgui.Separator()
        if imgui.Button("Reset Water Level") then
            reset()
        end

        imgui.SameLine()

        if imgui.Button("Reset All") then
            reset()
            setup()
            allWater = getAllWater()
            M.ocean = getOcean()
            M.sky = getSky()
        end

        imgui.Separator()
        tooltip("You can CTRL+Click on a slider to manually enter a value")

        if imgui.SliderFloat("Speed", ptrSpeed, 0, 1, "%.6f") then
            M.speed = ptrSpeed[0]
        end

        if imgui.SliderFloat("Rain Multiplier", ptrRainMultiplier, 0, 0.5, "%.6f") then
            M.rainMultiplier = ptrRainMultiplier[0]
        end

        if imgui.SliderFloat("Height Limit", ptrLimit, -500, 500, "%.6f") then
            M.limit = ptrLimit[0]
        end

        if imgui.SliderFloat("Water Level", ptrWaterLevel, -500, 500, "%.6f") then
            setWaterLevel(ptrWaterLevel[0])
        end
        imgui.Separator()

        local strHours = ""
        local strMinutes = ""
        if ptrHours[0] < 10 then strHours = "0" .. tostring(ptrHours[0]) else strHours = tostring(ptrHours[0]) end
        if ptrMinutes[0] < 10 then strMinutes = "0" .. tostring(ptrMinutes[0]) else strMinutes = tostring(ptrMinutes[0]) end

        floodui.textCenter("Environment | Time: " .. strHours .. ":" .. strMinutes)

        local updateTime = false
        if imgui.SliderFloat("Hours", ptrHours, 0, 24, "%.0f") then
            updateTime = true
        end

        if imgui.SliderFloat("Minutes", ptrMinutes, 0, 59, "%.0f") then
            updateTime = true
        end

        -- if imgui.SliderFloat4("Sky Color", ptrSkyCol, 0, 255, "%.0f") then
            
        -- end

        if updateTime then
            local hours = ptrHours[0]
            local minutes = ptrMinutes[0]
            local time = (((hours * 3600 + minutes * 60 + 0) / 86400) + 0.5) % 1
            core_environment.setTimeOfDay({time = time})
        end

        floodui.textCenter("Rain")

        if rainObj then
            if imgui.Button("Remove Rain") then
                windowMinSize = 470
                removeRainClicked = true

                rainObj:delete()
                local soundObj = findObject("rain_sound")
                if soundObj then
                    soundObj:delete()
                end
            end

            if imgui.SliderFloat("Rain Amount", ptrRainAmount, 0, 1000, "%.2f") then
                rainObj.numDrops = ptrRainAmount[0]
            end

            if imgui.SliderFloat("Rain Volume", ptrRainVolume, 0, 1, "%.2f") then
                setRainVolume(ptrRainVolume[0])
            end
            imgui.SameLine()
            tooltip("Sliding this will lag because I have to destroy and create a new sound source. Setting volume realtime doesn't work.")
        else
            if imgui.Button("Add Rain") then
                windowMinSize = 528
                
                -- create rain
                rainObj = createObject("Precipitation")
                rainObj.dataBlock = scenetree.findObject("rain_medium")
                rainObj.numDrops = 0
                rainObj.splashSize = 0
                rainObj.splashMS = 0
                rainObj.animateSplashes = 0
                rainObj.boxWidth = 16.0
                rainObj.boxHeight = 10.0
                rainObj.dropSize = 1.0
                rainObj.doCollision = true
                rainObj.hitVehicles = true
                rainObj.rotateWithCamVel = true
                rainObj.followCam = true
                rainObj.useWind = true
                rainObj.minSpeed = 0.4
                rainObj.maxSpeed = 0.5
                rainObj.minMass = 4
                rainObj.masMass = 5
                rainObj:registerObject('rain_coverage')

                setRainVolume(0)
            end
        end

        imgui.End()
    end
    floodui.popAll()
end

-- callbacks
local function onUpdate(dt)
    local mission = getMissionFilename()
    if not oldMission or oldMission ~= mission then
        if mission == "" then return end
        oldMission = mission

        setup()

        for _, water in pairs(allWater) do
            water.isRenderEnabled = true
        end

        allWater = getAllWater()
        M.ocean = getOcean()
        M.sky = getSky()

        setRainVolume(0)
        
        ptrWaterLevel[0] = getWaterLevel() or 0
    end

    if not mission then return end

    if M.showUI[0] then
        renderUI()
    end

    if not M.enabled or not M.ocean then return end
    hideCoveredWater()
    if bullettime.getPause() then return end

    local height = getWaterLevel()

    if M.floodWithRain then
        local amount = getRainAmount()
        local newHeight = height + amount * M.rainMultiplier * dt
        if not M.limitEnabled or newHeight <= M.limit then
            height = newHeight
        else
            height = M.limit
        end
    end

    if M.decreaseToInitial then
        local initialZ = initialWaterPosition:getColumn(3).z
        local curZ = ptrWaterLevel[0]

        local diff = tonumber(string.format("%.2f", initialZ - curZ))
        if diff <= 0 then diff = diff * -1 end

        if curZ > initialZ then
            local newZ = curZ - M.speed * dt
            ptrWaterLevel[0] = newZ
            setWaterLevel(newZ)
        elseif curZ < initialZ then
            local newZ = curZ + M.speed * dt
            ptrWaterLevel[0] = newZ
            setWaterLevel(newZ)
        else
            M.decreaseToInitial = false
            M.enabled = false
        end

        if diff >= -0.01 and diff <= 0.01 then -- somewhat helps if you change speed (which you shouldn't if you're decreasing)
            M.decreaseToInitial = false
            M.enabled = false
        end
    else
        if M.decrease then
            local target = height - M.speed
            
            if M.limitEnabled then
                if target > M.limit then
                    height = height - M.speed * dt
                end
            else
                height = height - M.speed * dt
            end
        else
            local target = height + M.speed

            if M.limitEnabled then
                if target < M.limit then
                    height = height + M.speed * dt
                end
            else
                height = height + M.speed * dt
            end
        end

        ptrWaterLevel[0] = height
        setWaterLevel(height)
    end
end

local function reload() -- useful for developing, best function in the game
    log("D", "reload", "Reloading...")
    extensions.reload("flood")
    extensions.reload("floodui")
end

local function onExtensionLoaded()
    log("D", "onExtensionLoaded", "Flood extension loaded")
end

local function onExtensionUnloaded()
    log("D", "onExtensionUnloaded", "Flood extension unloaded")
    local soundObj = findObject("rain_sound")
    local rainObj = findObject("rain_coverage", "Precipitation")
    if soundObj then
        setRainVolume(0)
    end
    if rainObj then
        rainObj.numDrops = 0
    end
    reset()
    setup()
end

M.resetWaterLevel = resetWaterLevel
M.resetAllWaterSources = resetAllWaterSources
M.reset = reset
M.getWaterLevel = getWaterLevel
M.setWaterLevel = setWaterLevel
M.reload = reload

M.onExtensionLoaded = onExtensionLoaded
M.onExtensionUnloaded = onExtensionUnloaded
M.onUpdate = onUpdate

return M